/* -*- mode: c++; c-basic-offset: 4; -*- */
#ifndef TRIGMGR_HH
#define TRIGMGR_HH

#include "SigFlag.hh"
#include "lmsg/MsgTypes.hh"
#include "AppServer.hh"
#include "lmsg/MsgHandler.hh"
#include "TrigMsg.hh"
#include "TrigStream.hh"
#include "TrigCtlTable.hh"
#include "TrigMgrLog.hh"

namespace html {
  class document;
  class text;
}

#include <vector>

//
//   The Following are bits for special purpose hacks set by the -hack 
//   argument field. These should not be used in general running.
//
//      HACK_SEQDISP  Rewrite the trigger disposition with a sequence number
#define HACK_SEQDISP 1

/**  The trigger manager (TrigMgr) class implements the trigger manager Service. 
  *  It receives messages containing the trigger or segment data from the 
  *  TrigClient client interface, collects and coalesces the data and writes 
  *  the collected data to ilwd (xsil/xml) data files vie oneormore trigger 
  *  streams (TrigStream class). Post-processing for the output files may 
  *  be specified when configuring the streams.
  *  \brief Trigger Manager implementation class.
  *  \author John Zweizig <john.zweizig@ligo.org>
  *  \version Last Modified: 22 May 2022.
  */
class TrigMgr : lmsg::AppServer {
public:

  /*   Initialize application.
   */
  TrigMgr(int argc, const char* argv[]);

  /*   Terminate application
   */
  ~TrigMgr(void);

  /*   Perform application processing.
   */
  int Application(void);

  /*   Close all output streams.
   *   @return Returns true if a configuration error occurred.
   */
  void closeStreams(void);

  /*   Configure output streams.
   *   @return Returns true if a configuration error occurred.
   */
  bool configStreams(const std::string& file);

  /*   Get latest write time.
   */
  Time getLastWrite(void) const;

  /*   Get number of triggers ready to be written.
   */
  int getNTrigs(void) const;

  /*   Handle client termination messages.
   */
  lmsg::error_type HandleClose(const lmsg::MsgHeader& h, const TMM_Close& msg);

  /*   Handle client registration messages
   *--------------------------------------*/
  lmsg::error_type HandleRegister(const lmsg::MsgHeader& h, const TMM_Register& msg);

  /*   Handle segment messages
   *--------------------------*/
  lmsg::error_type HandleSegment(const lmsg::MsgHeader& h, const TMM_Segment& msg);

  /*   Handle trigger messages
   *--------------------------*/
  lmsg::error_type HandleTrigger(const lmsg::MsgHeader& h, const TMM_Trigger& msg);

  /*   Handle rmNode messsages
   *--------------------------*/
  lmsg::error_type HandleRmNode(const lmsg::MsgHeader& h, const TMM_RemNode& msg);

  /*   Handle reConfig messsages
   *--------------------------*/
  lmsg::error_type HandleReconfig(const lmsg::MsgHeader& h, const TMM_Reconfigure& msg);

  void PurgeProcDB(void);

  void setTimer(Interval dt);

  void writeStats(void) const;
  void writeSegStats(void) const;
  void WriteDB(bool flush);

  void htmCreateLog(void);
  void htmAddLog(const trig::TrigBase& t, const trig::TrigProc& p);
  void htmWriteLog(void);

private:
  typedef std::vector<TrigStream*> StreamDB;
  typedef StreamDB::iterator stream_iter;

private:
    //----------------------------------  Trigger manager parameters
    bool     mEnableDB;
    bool     mEnableEpX;
    Interval mTimer;
    TrigStream::param mDefParam;
    std::string mOutDir;
    std::string mNotify;
    std::string mConfigFile;
    std::string mStreamConf;

    //----------------------------------  Trigger manager status
    SigFlag mTerm;
    SigFlag mForce;
    SigFlag mReconfigure;
    TrigCtlTable mCTable;
    Time mLastWrite;
    TrigMgrLog mTrigLog;

  //----------------------------------  Process, trigger, segment databases
    ProcessDB mProcDB;
    StreamDB  mStreamList;
};   //   class TrigMgr

#endif  //  TRIGMGR_HH
